﻿using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.CommonBot.Coroutines;
using Styx.Helpers;
using Styx.Plugins;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Numerics;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows.Media;

namespace UseFollowerAbility
{
    public class UseFollowerAbility : HBPlugin
    {
        private static Stopwatch logTimer = new Stopwatch();
        private static Stopwatch logInfoTimer = new Stopwatch();
        public static FaSettings S = new FaSettings();
        bool IsEnabled = false;
        private Composite combatHook;
        private static bool combatHookIsAoe;
        //static Coroutine currentRoutine;
        private void BotEvents_OnPulse(object sender, EventArgs e)
        {
            //infoLog("Hooks added: " + TreeHooks.Instance.Hooks.Count + " " + TreeHooks.Instance.Hooks["Combat_Main"].FirstOrDefault();

        }

        public async Task<bool> CheckAbilities()
        {
            //infoLog("Checking abilities");
            #region log stuff
            if (logInfoTimer.IsRunning && logInfoTimer.ElapsedMilliseconds > 500)
            {
                lastInfoMSG = "";
                logInfoTimer.Stop(); logInfoTimer.Reset();
            }
            if (logTimer.IsRunning && logTimer.ElapsedMilliseconds > 500)
            {
                lastCombatMSG = "";
                logTimer.Stop(); logTimer.Reset();
            }
            #endregion

            if (!StyxWoW.IsInGame || !StyxWoW.IsInWorld || Me == null || !Me.IsValid || !Me.IsAlive || Me.Mounted)
                return false;
            if (!Me.Combat)
                return false;

            // check the current player and make sure they have the ability before we even try anything
            if (PlayerHasRequiredAbility() && PlayerCanCastRequiredAbility())
            {
                //infoLog("PlayerHasRequiredAbility && PlayerCanCastRequiredAbility");
                // grab all enemies within 50 yards
                FindEnemies(Range: 50);

                // let's do stuff in close range for melee dps
                if (PlayerIsMeleeClass())
                {
                    //infoLog("PlayerIsMeleeClass()");
                    // we have 3 or more enemies within 8 yards
                    if (await CastMeleeAbilities(activeEnemies(Me.Location, 8f).Count() >= S.GeneralMobCount))
                    {
                        return true;
                    }
                }
                else
                {
                    // ranged targets don't need 8 yard radius
                    if (await CastRangedAbilities(activeEnemies(Me.Location, 40f).Count() >= S.GeneralMobCount))
                    {
                        return true;
                    }
                }
            }
            else
            {
                //RemoveCombatHook();
            }
            return false;
        }

        //private readonly List<Coroutine> _coroutines = new List<Coroutine>();

        public override void Pulse()
        {

        }

        #region Helper Methods

        #region Cast and Log

        /// <summary>
        /// Specifies the current target when casting spell and sleeps for lag
        /// </summary>
        /// <param name="Spell"></param>
        /// <param name="reqs"></param>
        /// <returns></returns>
        public async Task<bool> Cast(int Spell, Color newColor, bool reqs = true)
        {

            if (!CurrentTarget.IsValidCombatUnit())
                return false;
            if (!reqs)
            {
                //L.combatLog("Trying to cast: " + WoWSpell.FromId(Spell).Name + (String.IsNullOrEmpty(addLog) ? "" : " - " + addLog));
                return false;
            }

            if (!SpellManager.CanCast(WoWSpell.FromId(Spell), CurrentTarget, false, false, false)) //Should we check for if out currentTarget is moving? *Second false
                return false;
            if (!SpellManager.Cast(Spell, CurrentTarget))
                return false;
            //lastSpellCast = Spell;
            combatLog("^" + WoWSpell.FromId(Spell).Name, newColor);
            await CommonCoroutines.SleepForLagDuration();
            return true;
        }
        /// <summary>
        /// Does not specify a target when casting a spell and sleeps for lag
        /// </summary>
        /// <param name="Spell">The spell you wish to cast.</param>
        /// <param name="reqs">The requirements to cast the spell.</param>
        /// <returns></returns>
        public async Task<bool> GCD(int Spell, Color newColor, bool reqs = true)
        {

            if (!reqs)
            {
                //L.combatLog("Trying to cast: " + WoWSpell.FromId(Spell).Name + (String.IsNullOrEmpty(addLog) ? "" : " - " + addLog));
                return false;
            }
            //if (SpellManager.GlobalCooldown || !SpellManager.CanCast(Spell))
            //    return false;
            if (OnCooldown(Spell))
                return false;
            if (!SpellManager.Cast(Spell))
                return false;
            combatLog("*" + WoWSpell.FromId(Spell).Name, newColor);
            await CommonCoroutines.SleepForLagDuration();
            return true;
        }
        /// <summary>
        /// Determine from SPellFindResults if the spell is on cooldown or not.
        /// </summary>
        /// <param name="Spell"></param>
        /// <returns></returns>
        public bool OnCooldown(int Spell)
        {
            SpellFindResults Results;
            if (SpellManager.FindSpell(Spell, out Results)) { return Results.Override != null ? Results.Override.Cooldown : Results.Original.Cooldown; }
            return false;

        }
        /// <summary>
        /// Casts spell on the ground at current target's location.
        /// </summary>
        /// <param name="Spell">The spell you wish to cast.</param>
        /// <param name="reqs">The requirements to cast the spell.</param>
        /// <returns></returns>
        public async Task<bool> CastGround(int Spell, WoWUnit target, Color newColor, bool reqs = true, string addLog = "")
        {

            if (!reqs) { return false; }
            //L.combatLog("Trying to cast: " + WoWSpell.FromId(Spell).Name + (String.IsNullOrEmpty(addLog) ? "" : " - " + addLog));
            if (!target.IsValidCombatUnit()) { return false; }
            if (!SpellManager.CanCast(WoWSpell.FromId(Spell), target, false, false, false)) { return false; }

            if (await GCD(Spell, newColor, true)
                && !await Coroutine.Wait(1000, () => Me.CurrentPendingCursorSpell != null))
            {
                return false;
            }
            SpellManager.ClickRemoteLocation(target.Location);
            await CommonCoroutines.SleepForLagDuration();

            return true;
        }
        /// <summary>
        /// Casts spell on the ground at current target's location.
        /// </summary>
        /// <param name="Spell">The spell you wish to cast.</param>
        /// <param name="reqs">The requirements to cast the spell.</param>
        /// <returns></returns>
        public async Task<bool> CastGround(int Spell, Color newColor, bool reqs = true, string addLog = "")
        {

            if (!reqs) { return false; }
            if (await CastGround(Spell, CurrentTarget, newColor, reqs, addLog)) { return true; }
            return false;
        }

        public static void combatLog(string Message, System.Windows.Media.Color logColor, params object[] args)
        {
            if (Message == lastCombatMSG)
                return;

            if (CurrentTarget != null)
            {
                Logging.Write(logColor, "[UseFollowerAbility]: {0}" + string.Format($" on {CurrentTarget.SafeName} @ {CurrentTarget.HealthPercent.ToString("F2")}% at {CurrentTarget.Distance.ToString("F2")} yds with {Me.CurrentPower}/{Me.MaxPower} {CurrentPowerDescription}"), Message, args);
            }
            else
            {
                Logging.Write(logColor, "[UseFollowerAbility]: {0}" + string.Format($" on [no target] with {Me.CurrentPower}/{Me.MaxPower} {CurrentPowerDescription}"), Message, args);
            }


            lastCombatMSG = Message;
            logTimer.Start();
        }
        private static string lastInfoMSG;
        public static void infoLog(string Message, Color? differentLogColor = null)
        {
            if (Message == lastInfoMSG)
                return;


            Logging.Write(differentLogColor ?? Colors.White, string.Format($"[UseFollowerAbility]: {Message}"));

            lastInfoMSG = Message;
            logInfoTimer.Start();
        }
        #endregion

        #region Find Enemies
        public static IEnumerable<WoWUnit> activeEnemies(Vector3 fromLocation, double Range)
        {
            var Hostile = enemyCount;
            return Hostile != null ? Hostile.Where(x => x.Location.DistanceSquared(fromLocation) <= Range * Range) : null;
        }

        private static IEnumerable<WoWUnit> surroundingEnemies() { return ObjectManager.GetObjectsOfTypeFast<WoWUnit>(); }
        private static List<WoWUnit> enemyCount { get; set; }
        public static void FindEnemies(double Range)
        {

            enemyCount.Clear();
            foreach (var u in surroundingEnemies())
            {
                if (u == null || !u.IsValid)//<~ Use isUnitValid(u, Range) instead of the frist five of these first calls?
                    continue;
                if (!u.IsAlive || u.DistanceSqr > Range * Range)
                    continue;
                if (!u.Attackable || !u.CanSelect)
                    continue;
                if (u.IsFriendly)
                    continue;
                if (u.IsNonCombatPet && u.IsCritter)
                    continue;

                // make sure the unit is targeting something we care about.
                if (u.IsTargetingMeOrPet || u.IsTargetingMyPartyMember || u.IsTargetingMyRaidMember || u.SafeName == "Raider's Training Dummy")
                {
                    enemyCount.Add(u);
                }

            }
        }
        #endregion

        #region Has Abilities
        bool PlayerHasRequiredAbility(bool print = false)
        {
            int spellId = 0;
            bool hasAbility = false;
            SpellFindResults sfr;

            if (Me.Class == WoWClass.DeathKnight) { if (SpellManager.HasSpell(MeatHook)) { spellId = MeatHook; hasAbility = true; } }
            if (Me.Class == WoWClass.DemonHunter) { if (SpellManager.HasSpell(MothersCaress)) { spellId = MothersCaress; hasAbility = true; } }
            if (Me.Class == WoWClass.Druid) { if (SpellManager.HasSpell(NightmarishVisions)) { spellId = NightmarishVisions; hasAbility = true; } }
            if (Me.Class == WoWClass.Hunter) { if (SpellManager.HasSpell(EmmarelsAssault)) { spellId = EmmarelsAssault; hasAbility = true; } }
            if (Me.Class == WoWClass.Mage) { if (SpellManager.HasSpell(KalecgosFury)) { spellId = KalecgosFury; hasAbility = true; } }
            if (Me.Class == WoWClass.Paladin)
            {
                if (SpellManager.HasSpell(BloodVanguard)) { spellId = BloodVanguard; hasAbility = true; }
                if (SpellManager.HasSpell(JudgementsGaze)) { spellId = JudgementsGaze; hasAbility = true; }
            }
            if (Me.Class == WoWClass.Priest) { if (SpellManager.HasSpell(LightOfTheNaaru)) { spellId = LightOfTheNaaru; hasAbility = true; } }
            if (Me.Class == WoWClass.Rogue) { if (SpellManager.HasSpell(SealOfRavenholdt)) { spellId = SealOfRavenholdt; hasAbility = true; } }
            if (Me.Class == WoWClass.Shaman) { if (SpellManager.HasSpell(BoundlessQuintessence)) { spellId = BoundlessQuintessence; hasAbility = true; } }
            if (Me.Class == WoWClass.Warlock)
            {
                if (SpellManager.HasSpell(EternalBanishment)) { spellId = EternalBanishment; hasAbility = true; }
                if (SpellManager.HasSpell(SixSoulBag)) { spellId = SixSoulBag; hasAbility = true; }
            }
            if (Me.Class == WoWClass.Warrior)
            {
                if (SpellManager.HasSpell(WindsOfTheNorth)) { spellId = WindsOfTheNorth; hasAbility = true; }
                if (SpellManager.HasSpell(TitansWrath)) { spellId = TitansWrath; hasAbility = true; }
            }

            if (print && hasAbility)
            {
                if (SpellManager.FindSpell(spellId, out sfr))
                {

                    infoLog(string.Format($"{Me.SafeName} has follower abilty: {(sfr.Override != null ? sfr.Override.Name : sfr.Original.Name)}"));
                }
            }

            // default false
            return hasAbility;
        }

        bool PlayerCanCastRequiredAbility()
        {

            if (Me.Class == WoWClass.DeathKnight)
            {
                if (SpellManager.HasSpell(MeatHook) && SpellManager.CanCast(MeatHook)) { return true; }
            }
            if (Me.Class == WoWClass.DemonHunter)
            {
                if (SpellManager.HasSpell(MothersCaress) && SpellManager.CanCast(MothersCaress)) { return true; }
            }
            if (Me.Class == WoWClass.Druid)
            {
                if (SpellManager.HasSpell(NightmarishVisions) && SpellManager.CanCast(NightmarishVisions)) { return true; }
            }
            if (Me.Class == WoWClass.Hunter)
            {
                if (SpellManager.HasSpell(EmmarelsAssault) && SpellManager.CanCast(EmmarelsAssault)) { return true; }
            }
            if (Me.Class == WoWClass.Mage)
            {
                if (SpellManager.HasSpell(KalecgosFury) && SpellManager.CanCast(KalecgosFury)) { return true; }
            }
            if (Me.Class == WoWClass.Paladin)
            {
                if (SpellManager.HasSpell(BloodVanguard) && SpellManager.CanCast(BloodVanguard)) { return true; }
                if (SpellManager.HasSpell(JudgementsGaze) && SpellManager.CanCast(JudgementsGaze)) { return true; }
            }

            if (Me.Class == WoWClass.Priest)
            {
                if (SpellManager.HasSpell(LightOfTheNaaru) && SpellManager.CanCast(LightOfTheNaaru)) { return true; }
            }
            if (Me.Class == WoWClass.Rogue)
            {
                if (SpellManager.HasSpell(SealOfRavenholdt) && SpellManager.CanCast(SealOfRavenholdt)) { return true; }
            }
            if (Me.Class == WoWClass.Shaman)
            {
                if (SpellManager.HasSpell(BoundlessQuintessence) && SpellManager.CanCast(BoundlessQuintessence)) { return true; }
            }
            if (Me.Class == WoWClass.Warlock)
            {
                if (SpellManager.HasSpell(EternalBanishment) && SpellManager.CanCast(EternalBanishment)) { return true; }
                if (SpellManager.HasSpell(SixSoulBag) && SpellManager.CanCast(SixSoulBag)) { return true; }
            }
            if (Me.Class == WoWClass.Warrior)
            {
                if (SpellManager.HasSpell(WindsOfTheNorth) && SpellManager.CanCast(WindsOfTheNorth)) { return true; }
                if (SpellManager.HasSpell(TitansWrath) && SpellManager.CanCast(TitansWrath)) { return true; }
            }

            //infoLog("No follower ability ready");
            return false;
        }

        bool PlayerIsMeleeClass()
        {
            if (Me.Class == WoWClass.DeathKnight
                || Me.Class == WoWClass.DemonHunter
                || Me.Class == WoWClass.Rogue
                || Me.Class == WoWClass.Warrior
                || (Me.Class == WoWClass.Druid && (Me.Specialization == WoWSpec.DruidFeral || Me.Specialization == WoWSpec.DruidGuardian))
                || (Me.Class == WoWClass.Hunter && Me.Specialization == WoWSpec.HunterSurvival)
                || (Me.Class == WoWClass.Paladin && Me.Specialization != WoWSpec.PaladinHoly)
                || (Me.Class == WoWClass.Shaman && Me.Specialization == WoWSpec.ShamanEnhancement)
            )
            { return true; }

            return false;
        }
        bool UseBossEliteCheck()
        {
            return (S.UseOnlyOnBosses && CurrentTarget.IsBoss) || (S.UseOnlyOnElites && CurrentTarget.Elite) || (!S.UseOnlyOnBosses && !S.UseOnlyOnElites);
        }
        bool PlayerIsRangedClass()
        {
            return !PlayerIsMeleeClass();
        }
        /// <summary>
        /// These are cast within an 8 yard range since they are melee
        /// </summary>
        /// <param name="aoe"></param>
        /// <returns></returns>
        async Task<bool> CastMeleeAbilities(bool aoe = false)
        {
            if (Me.Class == WoWClass.DeathKnight) { if (await CastMeatHook(aoe)) { return true; } }
            if (Me.Class == WoWClass.DemonHunter) { if (await CastMothersCaress(aoe)) { return true; } }
            if (Me.Class == WoWClass.Rogue) { if (await CastSealOfRavenholdt()) { return true; } }
            if (Me.Class == WoWClass.Warrior) { if (await CastWindsOfTheNorth(aoe)) { return true; } }
            if (Me.Class == WoWClass.Warrior) { if (await CastTitansWrath(aoe)) { return true; } }
            if (Me.Class == WoWClass.Druid) { if (await CastNightmarishVisions()) { return true; } }
            if (Me.Class == WoWClass.Hunter) { if (await CastEmmarelsAssault(aoe)) { return true; } }
            if (Me.Class == WoWClass.Paladin) { if (await CastBloodVanguard(aoe)) { return true; } }
            if (Me.Class == WoWClass.Paladin) { if (await CastJudgementsGaze(aoe)) { return true; } }
            if (Me.Class == WoWClass.Shaman) { if (await CastBoundlessQuintessence()) { return true; } }

            return false;
        }

        async Task<bool> CastRangedAbilities(bool aoe = false)
        {
            if (Me.Class == WoWClass.Druid) { if (await CastNightmarishVisions()) { return true; } }
            if (Me.Class == WoWClass.Hunter) { if (await CastEmmarelsAssault(aoe)) { return true; } }
            if (Me.Class == WoWClass.Paladin) { if (await CastBloodVanguard(aoe)) { return true; } }
            if (Me.Class == WoWClass.Paladin) { if (await CastJudgementsGaze(aoe)) { return true; } }
            if (Me.Class == WoWClass.Shaman) { if (await CastBoundlessQuintessence()) { return true; } }
            if (Me.Class == WoWClass.Mage) { if (await CastKalecgosFury(aoe)) { return true; } }
            if (Me.Class == WoWClass.Priest) { if (await CastLightOfTheNaaru(aoe)) { return true; } }
            if (Me.Class == WoWClass.Warlock) { if (await CastEternalBanishment()) { return true; } }
            if (Me.Class == WoWClass.Warlock) { if (await CastSixSoulBag()) { return true; } }

            return false;
        }
        #region Cast Individual Spells
        async Task<bool> CastMeatHook(bool aoe)
        {

            if (aoe || (S.DkMeatHookLowHp && Me.HealthPercent <= S.DkMeatHookHpValue))
            {
                if (await Cast(MeatHook, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastMothersCaress(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await CastGround(MothersCaress, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }

        async Task<bool> CastNightmarishVisions()
        {

            if (UseBossEliteCheck())
            {
                if (await Cast(NightmarishVisions, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastEmmarelsAssault(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await CastGround(EmmarelsAssault, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastKalecgosFury(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await Cast(KalecgosFury, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }

        async Task<bool> CastBloodVanguard(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await Cast(BloodVanguard, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastJudgementsGaze(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await Cast(JudgementsGaze, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastLightOfTheNaaru(bool aoe)
        {

            if (aoe || S.PriestLightOfTheNaaruLowHp && Me.HealthPercent <= S.PriestLightOfTheNaaruHpValue)
            {
                if (await Cast(LightOfTheNaaru, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastSealOfRavenholdt()
        {

            if (UseBossEliteCheck() && CurrentTarget.HealthPercent >= S.RogueSealOfRavenholdtMinimumHpValue)
            {
                if (await Cast(SealOfRavenholdt, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastBoundlessQuintessence()
        {

            if (Me.HealthPercent <= S.ShamanBoundlessQuintessenceHpValue)
            {
                if (await Cast(BoundlessQuintessence, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastEternalBanishment()
        {

            if (UseBossEliteCheck())
            {
                if (await Cast(EternalBanishment, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastEredarTwins()
        {

            if (UseBossEliteCheck())
            {
                if (await Cast(EredarTwins, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastSixSoulBag()
        {

            if (UseBossEliteCheck())
            {
                if (await Cast(SixSoulBag, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastWindsOfTheNorth(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await Cast(WindsOfTheNorth, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }
        async Task<bool> CastTitansWrath(bool aoe)
        {

            if (aoe || UseBossEliteCheck())
            {
                if (await Cast(TitansWrath, Colors.Orange))
                {
                    return true;
                }
            }
            return false;
        }






        #endregion

        #endregion

        #region Hooks
        private void AddMeleeCombatHook(bool aoe)
        {
            // if the hook already matches and is present, then leave it alone.
            if (((combatHookIsAoe && aoe) || (!combatHookIsAoe && !aoe)) && combatHook != null) { infoLog("hook already added", Colors.Gray); return; }

            // remove hook if already there and aoe does not match
            if (combatHookIsAoe != aoe && combatHook != null)
            {
                //TreeHooks.Instance.RemoveHook(TreeHookExtension.LocationsNames.Combat_Main, combatHook);
                //RemoveCombatHook();
            }

            infoLog(string.Format($"Adding Melee Combat Hook, AoE: {aoe}", Colors.Green));

            combatHook = new ActionRunCoroutine(ctx => CastMeleeAbilities(aoe));
            //TreeHooks.Instance.AddHook(TreeHookExtension.LocationsNames.Combat_Main, combatHook);
            combatHookIsAoe = aoe;
        }
        private void AddRangedCombatHook(bool aoe)
        {
            // if the hook already matches and is present, then leave it alone.
            if (((combatHookIsAoe && aoe) || (!combatHookIsAoe && !aoe)) && combatHook != null) { infoLog("hook already added", Colors.Gray); return; }

            // remove hook if already there and aoe does not match
            if (combatHookIsAoe != aoe && combatHook != null)
            {
                //TreeHooks.Instance.RemoveHook(TreeHookExtension.LocationsNames.Combat_Main, combatHook);
                //RemoveCombatHook();
            }

            infoLog(string.Format($"Adding Melee Combat Hook, AoE: {aoe}", Colors.Green));

            combatHook = new ActionRunCoroutine(ctx => CastRangedAbilities(aoe));
            //TreeHooks.Instance.AddHook(TreeHookExtension.LocationsNames.Combat_Main, combatHook);
            combatHookIsAoe = aoe;
        }
        private void RemoveCombatHook()
        {
            if (combatHook != null)
            {
                //infoLog("Resume CR", Colors.Gray);
                TreeHooks.Instance.RemoveHook(TreeHookExtension.LocationsNames.Combat_Main, combatHook);
                combatHook = null;
                combatHookIsAoe = false;
                
                //TreeRoot.Resume();
            }
        }
        #endregion

        #endregion

        #region Static and overrides requirements
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static WoWUnit CurrentTarget { get { return StyxWoW.Me.CurrentTarget; } }
        public static string lastCombatMSG;
        private static string CurrentPowerDescription { get { return Me.PowerType.ToString().AddSpaces(); } }
        public override string Author { get { return "SpeshulK926"; } }
        public override string Name { get { return "Use Follower Ability"; } }
        public override Version Version { get { return new Version(1, 4); } }
        public override bool WantButton { get { return true; } }
        public override string ButtonText { get { return "UFA Settings"; } }
        public override void OnButtonPress()
        {
            SettingsForm frm = new SettingsForm();
            frm.Show();
            S.Save();
        }
        //private void ClearCoroutines()
        //{
        //    foreach (Coroutine ct in _coroutines)
        //        ct.Dispose();

        //    _coroutines.Clear();
        //}

       
        public override void OnEnable()
        {
            IsEnabled = true;
            S = new FaSettings();
            S.Load();
            BotEvents.OnPulse += BotEvents_OnPulse;
            BotEvents.OnBotStarted += BotEvents_OnBotStarted;
            BotEvents.OnBotStopped += BotEvents_OnBotStopped;
            enemyCount = new List<WoWUnit>();
            infoLog("Enabled!");
            if (!PlayerHasRequiredAbility(true))
                infoLog($"{Me.SafeName} does not have a follower ability!");

            PrintSettings();
            
        }

        private void BotEvents_OnBotStopped(EventArgs args)
        {
            RemoveCombatHook();
        }

        private void BotEvents_OnBotStarted(EventArgs args)
        {
            if (IsEnabled)
            {
                TreeHooks.Instance.InsertHook("Combat_Main", 0, combatHook ?? (combatHook = new ActionRunCoroutine(ctx => CheckAbilities())));
            }
            
        }

        public void PrintSettings()
        {
            Type type = S.GetType();
            PropertyInfo[] properties = type.GetProperties();

            foreach (PropertyInfo property in properties)
            {
                if (property.Name == "SettingsPath") { continue; }
                {
                    infoLog(string.Format($"{property.Name}: {property.GetValue(S, null)}"), Colors.Gray);
                }

            }
        }

        public override void OnDisable()
        {
            IsEnabled = false;
            //ClearCoroutines();
            enemyCount = null;
            infoLog("Disabled!", Colors.OrangeRed);
            RemoveCombatHook();
        }



        #endregion

        public const int
            MeatHook = 221462,
            MothersCaress = 223136,
            NightmarishVisions = 218533,
            EmmarelsAssault = 219698,
            KalecgosFury = 220680,
            BloodVanguard = 222621,
            JudgementsGaze = 221743,
            LightOfTheNaaru = 222113,
            SealOfRavenholdt = 222717,
            BoundlessQuintessence = 218277,
            EternalBanishment = 215977,
            EredarTwins = 216156,
            SixSoulBag = 215853,
            WindsOfTheNorth = 222935,
            TitansWrath = 223132,
            Done = 0;
    }

    static class Extensions
    {
        public static string AddSpaces(this string data)
        {
            Regex r = new Regex(@"(?!^)(?=[A-Z])");
            return r.Replace(data, " ");
        }
        public static bool HasAura(this Styx.WoWInternals.WoWObjects.LocalPlayer me, IEnumerable<int> auraList)
        {
            foreach (int aura in auraList)
            {
                if (me.HasAura(aura)) { return true; }
            }
            return false;
        }
        /// <summary>
        /// Returns a null-safe integer using an int.TryParse on the string.
        /// <para>return 0 if no good otherwise, will return the data</para>
        /// </summary>
        /// <param name="intData"></param>
        /// <returns></returns>
        public static uint SafeGetInt(this string intData)
        {
            try
            {
                uint safeInt = 0;
                if (uint.TryParse(intData, out safeInt))
                {
                    return safeInt;
                }
                return 0;
            }
            catch
            {
                return 0;
            }
        }

        public static bool IsValidCombatUnit(this WoWUnit Unit)
        {
            return Unit != null && Unit.IsValid && Unit.IsAlive && Unit.Attackable;
        }
        //public static bool IsInMeleeRange(this WoWUnit Unit, LocalPlayer me)
        //{
        //    return Unit.Distance < calculateMeleeRange(Unit, me);
        //}
        private static float calculateMeleeRange(WoWUnit Unit, LocalPlayer me)
        {
            return !me.GotTarget ? 0f : Math.Max(5f, me.CombatReach + 1.3333334f + Unit.CombatReach);
        }


    }

    public class FaSettings : Settings
    {
        public FaSettings()
            : base(Path.Combine(SettingsDirectory, "UseFollowerAbility"))
        { }
        [Setting, DefaultValue(3)]
        public int GeneralMobCount { get; set; }
        [Setting, DefaultValue(false)]
        public bool UseOnlyOnBosses { get; set; }
        [Setting, DefaultValue(false)]
        public bool UseOnlyOnElites { get; set; }
        [Setting, DefaultValue(true)]
        public bool DkMeatHookLowHp { get; set; }
        [Setting, DefaultValue(60)]
        public int DkMeatHookHpValue { get; set; }
        [Setting, DefaultValue(true)]
        public bool PriestLightOfTheNaaruLowHp { get; set; }
        [Setting, DefaultValue(60)]
        public int PriestLightOfTheNaaruHpValue { get; set; }
        [Setting, DefaultValue(90)]
        public int RogueSealOfRavenholdtMinimumHpValue { get; set; }
        [Setting, DefaultValue(60)]
        public int ShamanBoundlessQuintessenceHpValue { get; set; }

    }

    /// <summary> 
    /// Extension for TreeHook class. 
    /// </summary> 
    /// <remarks>Written by ZenLulz.</remarks> 
    public static class TreeHookExtension
    {

        public static void AddHook(this TreeHooks hook, LocationsNames location, Composite behavior)
        {
            hook.AddHook(location.ToString(), behavior);
        }

        public static void InsertHook(this TreeHooks hook, LocationsNames location, int index, Composite behavior)
        {
            hook.InsertHook(location.ToString(), index, behavior);
        }

        public static void RemoveHook(this TreeHooks hook, LocationsNames location, Composite behavior)
        {
            hook.RemoveHook(location.ToString(), behavior);
        }

        public static void ReplaceHook(this TreeHooks hook, LocationsNames location, Composite behavior)
        {
            hook.ReplaceHook(location.ToString(), behavior);
        }

        public static Dictionary<LocationsNames, List<CompositeListOperation>> GetHooks(this TreeHooks hook)
        {
            var dic = new Dictionary<LocationsNames, List<CompositeListOperation>>();

            hook.Hooks.ForEach(h => dic.Add((LocationsNames)Enum.Parse(typeof(LocationsNames), h.Key), h.Value));

            return dic;
        }

        // ReSharper disable InconsistentNaming 
        /// <summary> 
        /// Location Names for TreeHook. 
        /// </summary> 
        public enum LocationsNames
        {
            /// <summary> 
            /// The main combat root behavior location. Any combat logic is done within this hook, or child hooks. 
            /// </summary> 
            Combat_Main,
            /// <summary> 
            /// Combat: out of combat behavior hook location. Any "out of combat" routine related things are done here. 
            /// </summary> 
            Combat_OOC,
            /// <summary> 
            /// Routine: Rest Hook Location - This will call the Routine's Resting Functions such as eating and drinking. 
            /// </summary> 
            Routine_Rest,
            /// <summary> 
            /// Routine: Buff Hook Location - This will call the Routine's Buffing Functions such as 'Power Word: Fortitude', 'MotW' etc. 
            /// </summary> 
            Routine_Buff,
            /// <summary> 
            /// Combat: Pull Buff Hook Location - This will call the Routine's Pull Buff Functions such as Power Word: Shield right before pulling the target. 
            /// </summary> 
            Combat_PullBuff,
            /// <summary> 
            /// Combat: Pull Hook Location - This will call the Routine's Pull Functions such as Movement, Attacking, and LOS to get to start attacking the target. 
            /// </summary> 
            Combat_Pull,
            /// <summary> 
            /// Combat: Heal Hook Location - This will call the Routine's Heal Functions such as Healing other party members or even yourself. 
            /// </summary> 
            Combat_Heal,
            /// <summary> 
            /// Combat: Buff Hook Location - This will call the Routine's Combat Buff Functions such as 'Horn of Winter', 'Power Infusion' etc.. 
            /// </summary> 
            Combat_Buff,
            /// <summary> 
            /// Combat: Combat Only Hook Location - This will only call the Routine's Combat Function such as the main attack, will NOT call buffs, healing, resting, or pull. 
            /// </summary> 
            Combat_Only,
            /// <summary> 
            /// Death: Death Main Hook Location - This will call the Death Handling Functions such as Spirit Healer Res, Moving to body, and Moving to instance. 
            /// </summary> 
            Death_Main,
            /// <summary> 
            /// Death: Death Spirit Healer Hook Location - This will call the Death Spirit Healer Handling Functions that will talk to the Resser and accept.  
            /// Does not control the waiting after. 
            /// </summary> 
            Death_SpiritHealer,
            /// <summary> 
            /// Death: Death Corpse Run Hook Location - This will call the Death Corpse Run Handling Functions that will run or fly to your corpse including the death gryphons. 
            /// </summary> 
            Death_CorpseRun,
            /// <summary> 
            /// Death: Death Corpse Retrieval Hook Location - This will call the Death Corpse Retrieval Handling Functions that will find a safe spot, wait for the countdown,  
            /// move inside instance portal, and accept corpse resurrection. 
            /// </summary> 
            Death_CorpseRetrieval,
            /// <summary> 
            /// Loot: Loot Main Hook Location - This will call the Loots's Functions such as finding Mines/Herbs/Mobs to loot/Gameobject as well as dealing with looting blacklisting. 
            /// </summary> 
            Loot_Main,
            /// <summary> 
            /// Vendor: Vendor Main Hook Location - This will call the Vendors's Main Functions such as determining which vendors and mailboxes to go to, moving to the vendors,  
            /// and handling of the selling. 
            /// </summary> 
            Vendor_Main,
            /// <summary> 
            /// Vendor: Vendor Sell Hook Location - This will call the Vendors's Sell Functions to just sell items, will not handle opening of the Frame or closing it.  
            /// This will just sell items. 
            /// </summary> 
            Vendor_Sell,
            /// <summary> 
            /// Vendor: Vendor Repair Hook Location - This will call the Vendors's Repair Functions to just Repair items, will only repair items. 
            /// </summary> 
            Vendor_Repair,
            /// <summary> 
            /// Vendor: Vendor Mail Hook Location - This will call the Vendors's Mail Functions to just Mail items. Will not control the Frame. 
            /// </summary> 
            Vendor_Mail,
            /// <summary> 
            /// Vendor: Vendor Buy Hook Location - This will call the Vendors's Buy Functions to just Buy items. Will not control the Frame. 
            /// </summary> 
            Vendor_Buy,
            /// <summary> 
            /// Vendor: Vendor Train Hook Location - This will call the Vendors's Train Functions to any skill. This includes mount training and profession training.  
            /// Will not control the Frame. 
            /// </summary> 
            Vendor_Train
        }
        // ReSharper restore InconsistentNaming 
    }
}
